"""
   DeltaFi - Data transformation and enrichment platform

   Copyright 2022 DeltaFi Contributors <deltafi@deltafi.org>

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from deltafi.domain import *
from deltafi.exception import MissingMetadataException, MissingSourceMetadataException, ExpectedContentException, \
    MissingDomainException, MissingEnrichmentException


class DomainInput(NamedTuple):
    source_filename: str
    ingress_flow: str
    source_metadata: Dict[str, str]
    metadata: Dict[str, str]
    domains: Dict[str, Domain]

    def get_source_metadata(self, key: str) -> str:
        if key in self.source_metadata:
            return self.source_metadata[key]
        else:
            raise MissingSourceMetadataException

    def get_source_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.source_metadata:
            return self.source_metadata[key]
        else:
            return default

    def get_metadata(self, key: str):
        if key in self.metadata:
            return self.metadata[key]
        else:
            raise MissingMetadataException

    def get_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.metadata:
            return self.metadata[key]
        else:
            return default

    def has_domain(self, name: str) -> bool:
        return name in self.domains

    def domain(self, name: str) -> Domain:
        if not self.has_domain(name):
            raise MissingDomainException(name)
        return self.domains[name]


class EgressInput(NamedTuple):
    source_filename: str
    ingress_flow: str
    source_metadata: Dict[str, str]
    formatted_data: FormattedData

    def get_source_metadata(self, key: str) -> str:
        if key in self.source_metadata:
            return self.source_metadata[key]
        else:
            raise MissingSourceMetadataException

    def get_source_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.source_metadata:
            return self.source_metadata[key]
        else:
            return default


class EnrichInput(NamedTuple):
    source_filename: str
    ingress_flow: str
    source_metadata: Dict[str, str]
    content: List[Content]
    metadata: dict
    domains: Dict[str, Domain]
    enrichment: Dict[str, Domain]

    def get_source_metadata(self, key: str) -> str:
        if key in self.source_metadata:
            return self.source_metadata[key]
        else:
            raise MissingSourceMetadataException

    def get_source_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.source_metadata:
            return self.source_metadata[key]
        else:
            return default

    def has_content(self) -> bool:
        return len(self.content) > 0

    def content_at(self, index: int) -> Content:
        if len(self.content) < index + 1:
            raise ExpectedContentException(index, len(self.content))
        return self.content[index]

    def first_content(self):
        return self.content_at(0)

    def get_metadata(self, key: str):
        if key in self.metadata:
            return self.metadata[key]
        else:
            raise MissingMetadataException

    def get_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.metadata:
            return self.metadata[key]
        else:
            return default

    def has_domain(self, name: str) -> bool:
        return name in self.domains

    def domain(self, name: str) -> Domain:
        if not self.has_domain(name):
            raise MissingDomainException(name)
        return self.domains[name]

    def has_enrichment(self, name: str) -> bool:
        return name in self.enrichment

    def enrichment(self, name: str) -> Domain:
        if not self.has_enrichment(name):
            raise MissingEnrichmentException(name)
        return self.enrichment[name]


class FormatInput(NamedTuple):
    source_filename: str
    ingress_flow: str
    source_metadata: Dict[str, str]
    content: List[Content]
    metadata: dict
    domains: Dict[str, Domain]
    enrichment: Dict[str, Domain]

    def get_source_metadata(self, key: str) -> str:
        if key in self.source_metadata:
            return self.source_metadata[key]
        else:
            raise MissingSourceMetadataException

    def get_source_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.source_metadata:
            return self.source_metadata[key]
        else:
            return default

    def has_content(self) -> bool:
        return len(self.content) > 0

    def content_at(self, index: int) -> Content:
        if len(self.content) < index + 1:
            raise ExpectedContentException(index, len(self.content))
        return self.content[index]

    def first_content(self):
        return self.content_at(0)

    def get_metadata(self, key: str):
        if key in self.metadata:
            return self.metadata[key]
        else:
            raise MissingMetadataException

    def get_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.metadata:
            return self.metadata[key]
        else:
            return default

    def has_domain(self, name: str) -> bool:
        return name in self.domains

    def domain(self, name: str) -> Domain:
        if not self.has_domain(name):
            raise MissingDomainException(name)
        return self.domains[name]

    def has_enrichment(self, name: str) -> bool:
        return name in self.enrichment

    def enrichment(self, name: str) -> Domain:
        if not self.has_enrichment(name):
            raise MissingEnrichmentException(name)
        return self.enrichment[name]


class LoadInput(NamedTuple):
    source_filename: str
    ingress_flow: str
    source_metadata: Dict[str, str]
    content: List[Content]
    metadata: dict

    def get_source_metadata(self, key: str) -> str:
        if key in self.source_metadata:
            return self.source_metadata[key]
        else:
            raise MissingSourceMetadataException

    def get_source_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.source_metadata:
            return self.source_metadata[key]
        else:
            return default

    def has_content(self) -> bool:
        return len(self.content) > 0

    def content_at(self, index: int) -> Content:
        if len(self.content) < index + 1:
            raise ExpectedContentException(index, len(self.content))
        return self.content[index]

    def first_content(self):
        return self.content_at(0)

    def get_metadata(self, key: str):
        if key in self.metadata:
            return self.metadata[key]
        else:
            raise MissingMetadataException

    def get_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.metadata:
            return self.metadata[key]
        else:
            return default


class TransformInput(NamedTuple):
    source_filename: str
    ingress_flow: str
    source_metadata: Dict[str, str]
    content: List[Content]
    metadata: dict

    def get_source_metadata(self, key: str) -> str:
        if key in self.source_metadata:
            return self.source_metadata[key]
        else:
            raise MissingSourceMetadataException

    def get_source_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.source_metadata:
            return self.source_metadata[key]
        else:
            return default

    def has_content(self) -> bool:
        return len(self.content) > 0

    def content_at(self, index: int) -> Content:
        if len(self.content) < index + 1:
            raise ExpectedContentException(index, len(self.content))
        return self.content[index]

    def first_content(self):
        return self.content_at(0)

    def get_metadata(self, key: str):
        if key in self.metadata:
            return self.metadata[key]
        else:
            raise MissingMetadataException

    def get_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.metadata:
            return self.metadata[key]
        else:
            return default


class ValidateInput(NamedTuple):
    source_filename: str
    ingress_flow: str
    source_metadata: Dict[str, str]
    formatted_data: FormattedData

    def get_source_metadata(self, key: str) -> str:
        if key in self.source_metadata:
            return self.source_metadata[key]
        else:
            raise MissingSourceMetadataException

    def get_source_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.source_metadata:
            return self.source_metadata[key]
        else:
            return default



"""
   DeltaFi - Data transformation and enrichment platform

   Copyright 2022 DeltaFi Contributors <deltafi@deltafi.org>

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import json
import os
from os.path import isfile, join
from pathlib import Path

import pkg_resources
import requests
import sys
import threading
import time
import traceback
from typing import List

from deltafi.actioneventqueue import ActionEventQueue
from deltafi.configserver import ConfigServerClient
from deltafi.domain import Event
from deltafi.exception import ExpectedContentException, MissingDomainException, MissingEnrichmentException, \
    MissingSourceMetadataException, MissingMetadataException
from deltafi.logger import get_logger
from deltafi.result import ErrorResult
from deltafi.storage import ContentService


def _coordinates():
    return {
        'groupId': os.getenv('PROJECT_GROUP'),
        'artifactId': os.getenv('PROJECT_NAME'),
        'version': os.getenv('PROJECT_VERSION')
    }


def _setup_config_client(core_host):
    client = ConfigServerClient(f"http://{core_host}/config")
    client.sync()
    return client


def _setup_queue(config_client, max_connections):
    redis_url = os.getenv('DELTAFI_REDIS_MASTER_PORT')
    password = config_client.deltafi_common('redis.password')
    return ActionEventQueue(redis_url, max_connections, password)


def _setup_content_service(config_client):
    return ContentService(config_client.deltafi_common('minio.url'),
                          config_client.deltafi_common('minio.access-key'),
                          config_client.deltafi_common('minio.secret-key'))


class Plugin(object):
    def __init__(self, actions: List, description: str):
        self.actions = [action() for action in actions]
        self.description = description
        self.coordinates = _coordinates()

        self.core_host = os.getenv('DELTAFI_CORE_SERVICE_SERVICE_HOST')
        self.config_client = _setup_config_client(self.core_host)
        self.queue = _setup_queue(self.config_client, len(self.actions))
        self.content_service = _setup_content_service(self.config_client)

        if os.getenv('ACTIONS_HOSTNAME'):
            self.hostname = os.getenv('ACTIONS_HOSTNAME')
        elif os.getenv('HOSTNAME'):
            self.hostname = os.getenv('HOSTNAME')
        elif os.getenv('COMPUTERNAME'):
            self.hostname = os.getenv('COMPUTERNAME')
        else:
            self.hostname = 'UNKNOWN'

        self.logger = get_logger()

        self.logger.info(f"Initialized ActionRunner with actions {actions}")

    def action_name(self, action):
        return f"{self.coordinates['groupId']}.{action.__class__.__name__}"

    def _action_json(self, action):
        return {
            'name': self.action_name(action),
            'description': action.description,
            'type': action.action_type.name,
            'requiresDomains': action.requires_domains,
            'requiresEnrichments': action.requires_enrichments,
            'schema': action.param_class().schema()
        }

    def _register(self):
        flows_path = str(Path(os.path.dirname(os.path.abspath(sys.argv[0]))) / 'flows')
        flow_files = [f for f in os.listdir(flows_path) if isfile(join(flows_path, f))]

        variables = []
        if 'variables.json' in flow_files:
            flow_files.remove('variables.json')
            variables = json.load(open(join(flows_path, 'variables.json')))

        flows = [json.load(open(join(flows_path, f))) for f in flow_files]
        actions = [self._action_json(action) for action in self.actions]

        url = f"http://{self.core_host}/plugins"
        headers = {'Content-type': 'application/json'}
        registration_json = \
            {
                'pluginCoordinates': self.coordinates,
                'displayName': os.getenv('PROJECT_NAME'),
                'description': self.description,
                'actionKitVersion': pkg_resources.get_distribution('deltafi').version,
                'dependencies': [],
                'actions': actions,
                'variables': variables,
                'flowPlans': flows
            }

        self.logger.info(f"Registering plugin:\n{registration_json}")

        response = requests.post(url, headers=headers, json=registration_json)
        if not response.ok:
            self.logger.error(f"Failed to register plugin ({response.status_code}):\n{response.content}")
            exit(1)

    def run(self):
        self._register()
        for action in self.actions:
            threading.Thread(target=self._do_action, args=(action,)).start()

    def _do_action(self, action):
        action_logger = get_logger(self.action_name(action))

        action_logger.info(f"Listening on {self.action_name(action)}")
        while True:
            try:
                event_string = self.queue.take(self.action_name(action))
                event = Event.create(json.loads(event_string), self.hostname, self.content_service, action_logger)
                action_logger.debug(f"Processing event for did {event.context.did}")

                try:
                    result = action.execute(event)
                except ExpectedContentException as e:
                    result = ErrorResult(f"Action attempted to look up element {e.index + 1} (index {e.index}) from "
                                         f"content list of size {e.size}",
                                         f"{str(e)}\n{traceback.format_exc()}")
                except MissingDomainException as e:
                    result = ErrorResult(f"Action attempted to access domain {e.name}, which does not exist",
                                         f"{str(e)}\n{traceback.format_exc()}")
                except MissingEnrichmentException as e:
                    result = ErrorResult(f"Action attempted to access enrichment {e.name}, which does not exist",
                                         f"{str(e)}\n{traceback.format_exc()}")
                except MissingSourceMetadataException as e:
                    result = ErrorResult(f"Missing ingress metadata with key {e.key}",
                                         f"{str(e)}\n{traceback.format_exc()}")
                except MissingMetadataException as e:
                    result = ErrorResult(f"Missing metadata with key {e.key}",
                                         f"{str(e)}\n{traceback.format_exc()}")
                except BaseException as e:
                    result = ErrorResult(f"Action execution {type(e)} exception", f"{str(e)}\n{traceback.format_exc()}")

                response = {
                    'did': event.context.did,
                    'action': event.context.action_name,
                    'start': time.time(),
                    'stop': time.time(),
                    'type': result.result_type
                }
                if result.result_key is not None:
                    response[result.result_key] = result.response()

                self.queue.put('dgs', json.dumps(response))
            except BaseException as e:
                action_logger.error(f"Unexpected {type(e)} error: {str(e)}\n{traceback.format_exc()}")
                time.sleep(1)



"""
   DeltaFi - Data transformation and enrichment platform

   Copyright 2022 DeltaFi Contributors <deltafi@deltafi.org>

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from logging import Logger
from typing import Dict, List, NamedTuple

from deltafi.storage import ContentService, ContentReference


def _keyval_to_metadata(keyval: List[Dict[str, str]]):
    return {item['key']: item['value'] for item in keyval}


def _metadata_to_keyval(metadata: Dict[str, str]):
    return [{'key': k, 'value': v} for (k, v) in metadata.items()]


class Content(NamedTuple):
    name: str
    metadata: Dict[str, str]
    content_reference: ContentReference

    def json(self):
        return {
            'name': self.name,
            'contentReference': self.content_reference.json(),
            'metadata': _metadata_to_keyval(self.metadata)
        }

    @classmethod
    def from_dict(cls, content: dict):
        name = content['name']
        metadata = _keyval_to_metadata(content.get('metadata', []))
        content_reference = ContentReference.from_dict(content['contentReference'])
        return Content(name=name,
                       metadata=metadata,
                       content_reference=content_reference)


class Context(NamedTuple):
    did: str
    action_name: str
    ingress_flow: str
    egress_flow: str
    system: str
    hostname: str
    content_service: ContentService
    logger: Logger

    @classmethod
    def create(cls, context: dict, hostname: str, content_service: ContentService, logger: Logger):
        did = context['did']
        action_name = context['name']
        ingress_flow = context['ingressFlow']
        if 'egressFlow' in context:
            egress_flow = context['egressFlow']
        else:
            egress_flow = None
        system = context['systemName']
        return Context(did=did,
                       action_name=action_name,
                       ingress_flow=ingress_flow,
                       egress_flow=egress_flow,
                       system=system,
                       hostname=hostname,
                       content_service=content_service,
                       logger=logger)


class Domain(NamedTuple):
    name: str
    value: str
    media_type: str

    @classmethod
    def from_dict(cls, domain: dict):
        name = domain['name']
        value = domain['value']
        media_type = domain['mediaType']
        return Domain(name=name,
                      value=value,
                      media_type=media_type)


class FormattedData(NamedTuple):
    filename: str
    format_action: str
    content_reference: ContentReference
    metadata: Dict[str, str]

    @classmethod
    def from_dict(cls, formatted_data: dict):
        filename = formatted_data['filename']
        format_action = formatted_data['formatAction']
        content_reference = ContentReference.from_dict(formatted_data['contentReference'])
        metadata = _keyval_to_metadata(formatted_data.get('metadata', []))
        return FormattedData(filename=filename,
                             format_action=format_action,
                             content_reference=content_reference,
                             metadata=metadata)


class ProtocolLayer(NamedTuple):
    action: str
    content: List[Content]
    metadata: Dict[str, str]

    @classmethod
    def from_dict(cls, layer: dict):
        action = layer['action']
        content = [Content.from_dict(item) for item in layer['content']]
        metadata = _keyval_to_metadata(layer.get('metadata', []))
        return ProtocolLayer(action=action,
                             content=content,
                             metadata=metadata)


class SourceInfo(NamedTuple):
    filename: str
    flow: str
    metadata: Dict[str, str]

    @classmethod
    def from_dict(cls, source_info: dict):
        filename = source_info['filename']
        flow = source_info['flow']
        metadata = _keyval_to_metadata(source_info.get('metadata', []))
        return SourceInfo(filename=filename,
                          flow=flow,
                          metadata=metadata)


class DeltaFile(NamedTuple):
    did: str
    source_info: SourceInfo
    protocol_stack: List[ProtocolLayer]
    domains: List[Domain]
    indexed_metadata: Dict[str, str]
    enrichment: List[Domain]
    formatted_data: FormattedData

    @classmethod
    def from_dict(cls, delta_file: dict):
        did = delta_file['did']
        source_info = SourceInfo.from_dict(delta_file['sourceInfo'])
        protocol_stack = [ProtocolLayer.from_dict(layer) for layer in delta_file['protocolStack']]
        domains = [Domain.from_dict(domain) for domain in delta_file['domains']]
        indexed_metadata = delta_file['indexedMetadata']
        enrichment = [Domain.from_dict(domain) for domain in delta_file['enrichment']]
        if len(delta_file['formattedData']) > 0:
            formatted_data = FormattedData.from_dict(delta_file['formattedData'][0])
        else:
            formatted_data = None
        return DeltaFile(did=did,
                         source_info=source_info,
                         protocol_stack=protocol_stack,
                         domains=domains,
                         indexed_metadata=indexed_metadata,
                         enrichment=enrichment,
                         formatted_data=formatted_data)


class Event(NamedTuple):
    delta_file: DeltaFile
    context: Context
    params: dict

    @classmethod
    def create(cls, event: dict, hostname: str, content_service: ContentService, logger: Logger):
        delta_file = DeltaFile.from_dict(event['deltaFile'])
        context = Context.create(event['actionContext'], hostname, content_service, logger)
        params = event['actionParams']
        return Event(delta_file, context, params)



"""
   DeltaFi - Data transformation and enrichment platform

   Copyright 2022 DeltaFi Contributors <deltafi@deltafi.org>

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import io
import minio
from typing import List, NamedTuple
from urllib.parse import urlparse
import uuid

BUCKET = 'storage'


class Segment(NamedTuple):
    uuid: str
    offset: int
    size: int
    did: str

    def json(self):
        return {
            'uuid': str(self.uuid),
            'offset': self.offset,
            'size': self.size,
            'did': self.did
        }

    @classmethod
    def from_dict(cls, segment: dict):
        s_uuid = segment['uuid']
        offset = segment['offset']
        size = segment['size']
        did = segment['did']
        return Segment(uuid=s_uuid,
                       offset=offset,
                       size=size,
                       did=did)

    def id(self):
        return f"{self.did}/{self.uuid}"


class ContentReference(NamedTuple):
    segments: List[Segment]
    media_type: str

    def json(self):
        return {
            'segments': [segment.json() for segment in self.segments],
            'mediaType': self.media_type
        }

    @classmethod
    def from_dict(cls, content_reference: dict):
        segments = [Segment.from_dict(segment) for segment in content_reference['segments']]
        media_type = content_reference['mediaType']
        return ContentReference(segments=segments,
                                media_type=media_type)


class ContentService:
    def __init__(self, url, access_key, secret_key):
        parsed = urlparse(url)
        self.minio_client = minio.Minio(
            f"{parsed.hostname}:{str(parsed.port)}",
            access_key=access_key,
            secret_key=secret_key,
            secure=False
        )

        found = self.minio_client.bucket_exists(BUCKET)
        if not found:
            raise RuntimeError(f"Minio bucket {BUCKET} not found")

    def get_bytes(self, content_reference: ContentReference):
        return b"".join([self.minio_client.get_object(BUCKET, segment.id(), segment.offset,
                                            segment.size).read() for segment in content_reference.segments])

    def get_str(self, content_reference: ContentReference):
        return self.get_bytes(content_reference).decode('utf-8')

    def put_bytes(self, did, bytes_data, media_type):
        segment = Segment(uuid=str(uuid.uuid4()),
                          offset=0,
                          size=len(bytes_data),
                          did=did)
        content_reference = ContentReference(segments=[segment],
                                             media_type=media_type)
        self.minio_client.put_object(BUCKET, segment.id(), io.BytesIO(bytes_data), len(bytes_data))
        return content_reference

    def put_str(self, did, string_data, media_type):
        return self.put_bytes(did, string_data.encode('utf-8'), media_type)



"""
   DeltaFi - Data transformation and enrichment platform

   Copyright 2022 DeltaFi Contributors <deltafi@deltafi.org>

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import os
import re
import sys
from typing import Any

import requests

from deltafi.logger import get_logger


def _make_key(property_set, key):
    return property_set + ":" + key


class ConfigServerClient:
    def __init__(self, url: str):
        self.url = url + "/application/default"
        self.config_map = {}
        self.logger = get_logger()

    def sync(self) -> None:
        """Query and save properties from config server"""
        try:
            response = requests.get(self.url)
            if response.ok:
                output = response.json()
                self.__parse__(output)
            else:

                self.logger.error(f"Failed to retrieve the configuration from {self.url} ({response.status_code}):\n"
                                  f"{response.content}")
                sys.exit(1)

        except RuntimeError as e:
            self.logger.error(f"Failed to retrieve the configuration from {self.url}:\n{e}")
            sys.exit(1)

    def deltafi_common(self, key):
        return self._lookup("deltafi-common", key)

    def action_kit(self, key):
        return self._lookup("action-kit", key)

    def _lookup(self, property_set, key) -> Any:
        compound_key = _make_key(property_set, key)
        value = self.config_map.get(compound_key)
        if value is not None:
            pattern = "\\$\\{(.*)\\}"
            result = re.match(pattern, value)
            if result:
                env_name = result.group(1)
                env_value = os.getenv(env_name)
                if not env_value:
                    self.logger.error("ENV not found: " + env_name)
                return env_value
        return value

    def __parse__(self, output):
        prop_sources = output.get("propertySources")
        self.config_map.clear()
        for src in prop_sources:
            name = src.get('name').split()[0]
            properties = src.get('source')
            keys = properties.keys()
            for key in keys:
                compound_key = _make_key(name, key)
                # Preserve override value, if present
                # TODO: This might not work if the override is null
                if self.config_map.get(compound_key) is None:
                    self.config_map.update({compound_key: properties.get(key)})



"""
   DeltaFi - Data transformation and enrichment platform

   Copyright 2022 DeltaFi Contributors <deltafi@deltafi.org>

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from datetime import datetime
import json_logging
import logging
import sys


def get_logger(name: str = None) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))

    if name is not None:
        logger = logging.LoggerAdapter(logger, dict(action=name))

    return logger


def _sanitize_log_msg(record):
    return record.getMessage().replace('\n', '_').replace('\r', '_').replace('\t', '_')


class JSONLogFormatter(json_logging.JSONLogFormatter):

    def _format_log_object(self, record, request_util):
        utcnow = datetime.utcnow()

        json_log_object = {
            'timestamp': json_logging.util.iso_time_format(utcnow),
            'message': _sanitize_log_msg(record),
            'loggerName': record.name,
            'threadName': record.threadName,
            'level': record.levelname,
            'module': record.module,
            'line_no': record.lineno
        }

        if hasattr(record, 'action'):
            json_log_object['action'] = record.action

        if hasattr(record, 'props'):
            json_log_object.update(record.props)

        if record.exc_info or record.exc_text:
            json_log_object.update(self.get_exc_fields(record))

        return json_log_object


json_logging.init_non_web(custom_formatter=JSONLogFormatter, enable_json=True)



"""
   DeltaFi - Data transformation and enrichment platform

   Copyright 2022 DeltaFi Contributors <deltafi@deltafi.org>

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from deltafi.actiontype import ActionType
from deltafi.domain import Context
from deltafi.input import DomainInput, EgressInput, EnrichInput, FormatInput, LoadInput, TransformInput, ValidateInput
from deltafi.result import *
from abc import ABC, abstractmethod
from pydantic import BaseModel


class Action(ABC):
    def __init__(self, action_type: ActionType, description: str, requires_domains: List[str],
                 requires_enrichments: List[str]):
        self.action_type = action_type
        self.description = description
        self.requires_domains = requires_domains
        self.requires_enrichments = requires_enrichments

    @abstractmethod
    def execute(self, event):
        pass

    def param_class(self):
        return BaseModel

    def validate_type(self, result, types):
        if not isinstance(result, types):
            raise ValueError(f"{self.__class__.__name__} must return one of "
                             f"{[result_type.__name__ for result_type in list(types)]} "
                             f"but a {result.__class__.__name__} was returned")


class DomainAction(Action):
    def __init__(self, description: str, requires_domains: List[str]):
        super().__init__(ActionType.DOMAIN, description, requires_domains, [])

    def execute(self, event):
        domain_input = DomainInput(source_filename=event.delta_file.source_info.filename,
                                   ingress_flow=event.delta_file.source_info.flow,
                                   source_metadata=event.delta_file.source_info.metadata,
                                   metadata=event.delta_file.protocol_stack[-1].metadata,
                                   domains={domain.name: domain for domain in event.delta_file.domains})
        result = self.domain(event.context, self.param_class().parse_obj(event.params), domain_input)
        self.validate_type(result, (DomainResult, ErrorResult))
        return result

    @abstractmethod
    def domain(self, context: Context, params: BaseModel, domain_input: DomainInput):
        pass


class EgressAction(Action):
    def __init__(self, description: str):
        super().__init__(ActionType.EGRESS, description, [], [])

    def execute(self, event):
        egress_input = EgressInput(source_filename=event.delta_file.source_info.filename,
                                   ingress_flow=event.delta_file.source_info.flow,
                                   source_metadata=event.delta_file.source_info.metadata,
                                   formatted_data=event.delta_file.formatted_data)
        result = self.egress(event.context, self.param_class().parse_obj(event.params), egress_input)
        self.validate_type(result, (EgressResult, ErrorResult, FilterResult))
        return result

    @abstractmethod
    def egress(self, context: Context, params: BaseModel, egress_input: EgressInput):
        pass


class EnrichAction(Action):
    def __init__(self, description: str, requires_domains: List[str], requires_enrichments: List[str]):
        super().__init__(ActionType.ENRICH, description, requires_domains, requires_enrichments)

    def execute(self, event):
        enrich_input = EnrichInput(source_filename=event.delta_file.source_info.filename,
                                   ingress_flow=event.delta_file.source_info.flow,
                                   source_metadata=event.delta_file.source_info.metadata,
                                   content=event.delta_file.protocol_stack[-1].content,
                                   metadata=event.delta_file.protocol_stack[-1].metadata,
                                   domains={domain.name: domain for domain in event.delta_file.domains},
                                   enrichment={domain.name: domain for domain in event.delta_file.enrichment})
        result = self.enrich(event.context, self.param_class().parse_obj(event.params), enrich_input)
        self.validate_type(result, (EnrichResult, ErrorResult))
        return result

    @abstractmethod
    def enrich(self, context: Context, params: BaseModel, enrich_input: EnrichInput):
        pass


class FormatAction(Action):
    def __init__(self, description: str, requires_domains: List[str], requires_enrichments: List[str]):
        super().__init__(ActionType.FORMAT, description, requires_domains, requires_enrichments)

    def execute(self, event):
        format_input = FormatInput(source_filename=event.delta_file.source_info.filename,
                                   ingress_flow=event.delta_file.source_info.flow,
                                   source_metadata=event.delta_file.source_info.metadata,
                                   content=event.delta_file.protocol_stack[-1].content,
                                   metadata=event.delta_file.protocol_stack[-1].metadata,
                                   domains={domain.name: domain for domain in event.delta_file.domains},
                                   enrichment={domain.name: domain for domain in event.delta_file.enrichment})
        result = self.format(event.context, self.param_class().parse_obj(event.params), format_input)
        self.validate_type(result, (FormatResult, FormatManyResult, ErrorResult, FilterResult))
        return result

    @abstractmethod
    def format(self, context: Context, params: BaseModel, format_input: FormatInput):
        pass


class LoadAction(Action):
    def __init__(self, description: str):
        super().__init__(ActionType.LOAD, description, [], [])

    def execute(self, event):
        load_input = LoadInput(source_filename=event.delta_file.source_info.filename,
                               ingress_flow=event.delta_file.source_info.flow,
                               source_metadata=event.delta_file.source_info.metadata,
                               content=event.delta_file.protocol_stack[-1].content,
                               metadata=event.delta_file.protocol_stack[-1].metadata)
        result = self.load(event.context, self.param_class().parse_obj(event.params), load_input)
        self.validate_type(result, (LoadResult, ErrorResult, FilterResult, SplitResult))
        return result

    @abstractmethod
    def load(self, context: Context, params: BaseModel, load_input: LoadInput):
        pass


class TransformAction(Action):
    def __init__(self, description: str):
        super().__init__(ActionType.TRANSFORM, description, [], [])

    def execute(self, event):
        transform_input = TransformInput(source_filename=event.delta_file.source_info.filename,
                                         ingress_flow=event.delta_file.source_info.flow,
                                         source_metadata=event.delta_file.source_info.metadata,
                                         content=event.delta_file.protocol_stack[-1].content,
                                         metadata=event.delta_file.protocol_stack[-1].metadata)
        result = self.transform(event.context, self.param_class().parse_obj(event.params), transform_input)
        self.validate_type(result, (TransformResult, ErrorResult, FilterResult))
        return result

    @abstractmethod
    def transform(self, context: Context, params: BaseModel, transform_input: TransformInput):
        pass


class ValidateAction(Action):
    def __init__(self, description: str):
        super().__init__(ActionType.VALIDATE, description, [], [])

    def execute(self, event):
        validate_input = ValidateInput(source_filename=event.delta_file.source_info.filename,
                                       ingress_flow=event.delta_file.source_info.flow,
                                       source_metadata=event.delta_file.source_info.metadata,
                                       formatted_data=event.delta_file.formatted_data)
        result = self.validate(event.context, self.param_class().parse_obj(event.params), validate_input)
        self.validate_type(result, (ValidateResult, ErrorResult, FilterResult))
        return result

    @abstractmethod
    def validate(self, context: Context, params: BaseModel, validate_input: ValidateInput):
        pass



"""
   DeltaFi - Data transformation and enrichment platform

   Copyright 2022 DeltaFi Contributors <deltafi@deltafi.org>

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import abc
from typing import Dict, List

from deltafi.domain import Content
from deltafi.storage import ContentReference


class Result:
    __metaclass__ = abc.ABCMeta

    def __init__(self, result_key, result_type):
        self.result_key = result_key
        self.result_type = result_type

    @abc.abstractmethod
    def response(self):
        pass


class DomainResult(Result):
    def __init__(self):
        super().__init__('domain', 'DOMAIN')
        self.indexed_metadata = {}

    def index_metadata(self, key: str, value: str):
        self.indexed_metadata[key] = value
        return self

    def response(self):
        return {
            'indexedMetadata': self.indexed_metadata
        }


class EgressResult(Result):
    def __init__(self):
        super().__init__(None, 'EGRESS')

    def response(self):
        return None


class EnrichResult(Result):
    def __init__(self):
        super().__init__('enrich', 'ENRICH')
        self.enrichments = []
        self.indexed_metadata = {}

    def enrich(self, name: str, value: str, media_type: str):
        self.enrichments.append({
            'name': name,
            'value': value,
            'mediaType': media_type
        })
        return self

    def index_metadata(self, key: str, value: str):
        self.indexed_metadata[key] = value
        return self

    def response(self):
        return {
            'enrichments': self.enrichments,
            'indexedMetadata': self.indexed_metadata
        }


class ErrorResult(Result):
    def __init__(self, cause: str, context: str):
        super().__init__('error', 'ERROR')
        self.cause = cause
        self.context = context

    def response(self):
        return {
            'cause': self.cause,
            'context': self.context
        }


class FilterResult(Result):
    def __init__(self, message: str):
        super().__init__('filter', 'FILTER')
        self.message = message

    def response(self):
        return {
            'message': self.message
        }


class FormatResult(Result):
    def __init__(self, filename: str, content_reference: ContentReference):
        super().__init__('format', 'FORMAT')
        self.filename = filename
        self.content_reference = content_reference
        self.metadata = []

    def add_metadata(self, key: str, value: str):
        self.metadata.append({'key': key, 'value': value})
        return self

    def response(self):
        return {
            'filename': self.filename,
            'contentReference': self.content_reference.json(),
            'metadata': self.metadata
        }


class FormatManyResult(Result):
    def __init__(self):
        super().__init__('formatMany', 'FORMAT_MANY')
        self.format_results = []

    def add_format_result(self, format_result: FormatResult):
        self.format_results.append(format_result)
        return self

    def response(self):
        return [format_result.response() for format_result in self.format_results]


class LoadResult(Result):
    def __init__(self):
        super().__init__('load', 'LOAD')
        self.content = []
        self.metadata = []
        self.domains = []

    def add_content(self, name: str, content_reference: ContentReference):
        content = Content(name=name, metadata={}, content_reference=content_reference)
        self.content.append(content)
        return self

    def add_metadata(self, key: str, value: str):
        self.metadata.append({'key': key, 'value': value})
        return self

    def add_domain(self, name: str, value: str, media_type: str):
        self.domains.append({
            'name': name,
            'value': value,
            'mediaType': media_type})
        return self

    def response(self):
        return {
            'domains': self.domains,
            'protocolLayer': {
                'content': [content.json() for content in self.content],
                'metadata': self.metadata
            }
        }


class SplitResult(Result):
    class SplitChild:
        def __init__(self, filename: str, flow: str, metadata: Dict[str, str], content: List[Content]):
            self.filename = filename
            self.flow = flow
            self.metadata = metadata
            self.content = content

        def json(self):
            return {
                'sourceInfo': {
                    'filename': self.filename,
                    'flow': self.flow,
                    'metadata': [{'key': k, 'value': v} for (k, v) in self.metadata.items()]
                },
                'content': [content.json() for content in self.content]
            }

    def __init__(self):
        super().__init__('split', 'SPLIT')
        self.children = []

    def add_child(self, filename: str, flow: str, metadata: Dict[str, str], content: List[Content]):
        child = SplitResult.SplitChild(filename, flow, metadata, content)
        self.children.append(child)

    def response(self):
        return [child.json() for child in self.children]


class TransformResult(Result):
    def __init__(self):
        super().__init__('transform', 'TRANSFORM')
        self.content = []
        self.metadata = []

    def add_content(self, name: str, content_reference: ContentReference):
        content = Content(name=name, metadata={}, content_reference=content_reference)
        self.content.append(content)
        return self

    def add_metadata(self, key: str, value: str):
        self.metadata.append({'key': key, 'value': value})
        return self

    def response(self):
        return {
            'protocolLayer': {
                'content': [content.json() for content in self.content],
                'metadata': self.metadata
            }
        }


class ValidateResult(Result):
    def __init__(self):
        super().__init__(None, 'VALIDATE')

    def response(self):
        return None
